#!/usr/bin/env python3
# survey script v1.0.3
import base64
import configparser
import getpass
import hashlib
import json
import os
import re
import shutil
import subprocess
import sys
import unicodedata
import uuid

from concurrent.futures import ThreadPoolExecutor
from datetime import datetime
from tempfile import TemporaryDirectory


try:
    from pyVmomi import vim
    from pyVim.connect import SmartConnectNoSSL, Disconnect
except ModuleNotFoundError:
    try:
        # some versions are lowercase
        from pyvmomi import vim
        from pyvim.connect import SmartConnectNoSSL, Disconnect
    except ModuleNotFoundError:
        print("Failed to import pyVmomi")

VERSION = "1.0.3"


class SurveyError(ValueError):
    pass


class VersionError(ValueError):
    pass


class ConnectionError(ValueError):
    pass


def is_in_development():
    return os.getenv("RUNTIME_ENV", "").lower() in ["dev", "development", "testing"]


def sanitize(value):
    value_str = str(value)
    value_ascii = unicodedata.normalize("NFKD", value_str).encode("ascii", "ignore").decode("ascii")
    value_nospaces = re.sub(r"[^\w\s-]", "", value_ascii).strip()
    return re.sub(r"[-\s]+", "-", value_nospaces)


def collect_migration_data(check_license=True):
    try:
        surveyor = Surveyor()
        surveyor.check_license = check_license

        with TemporaryDirectory() as target_dir:
            serial = surveyor.collect_system_configuration(target_dir)
            # Set restrictive permissions on every file in the directory before archiving
            for root, subdirs, files in os.walk(target_dir):
                for dir_name in subdirs:
                    os.chmod(os.path.join(root, dir_name), 0o700)
                for file_name in files:
                    os.chmod(os.path.join(root, file_name), 0o600)

            sanitized_serial = sanitize(serial)
            timestamp = datetime.now().strftime("%Y-%m-%d_%H-%M-%S")
            archive_name = "migration_data_{sanitized_serial}_{timestamp}".format(
                sanitized_serial=sanitized_serial, timestamp=timestamp
            )
            output_locations = surveyor.get_zpool_roots()

            archive_name = shutil.make_archive(
                base_name=os.path.join(output_locations[0], archive_name),
                format="gztar",
                root_dir=target_dir,
            )

        # Set file permission to read/write for owner only for the archive itself
        os.chmod(archive_name, 0o600)
        print("Generated: {archive_name}".format(archive_name=archive_name))

        for location in output_locations[1:]:
            archive_copy = shutil.copy2(archive_name, location)
            os.chmod(archive_name, 0o600)
            print("Generated: {archive_copy}".format(archive_copy=archive_copy))

        return archive_name
    except KeyboardInterrupt:
        print("Received [Ctrl-C], exiting.")
        raise SystemExit(1)
    except Exception as e:
        if is_in_development():
            raise
        else:
            print(
                "Failed to collect data error={error} message={message}".format(error=type(e).__name__, message=str(e))
            )
            raise SystemExit(1)


class Surveyor:
    def __init__(self):
        self._hypervisors = None
        self._serial = None
        self._version = None
        self._pools = None
        self.check_license = True
        self._datasets_reporting_data = None
        self._vm_reporting_data = None

    def collect_system_configuration(self, target_dir):
        print("Collecting SynetoOS4 migration metadata")

        self._pools = self.find_zfs_pools()
        if not self._pools:
            raise SurveyError("Found zero storage pools")

        # Define a list of tuples with the description, function to call, and the filename
        metadata_items = [
            ("Survey", self.find_survey_script_version, "survey.json"),
            ("Firmware Version", self.find_firmware_version, "firmware-version.json"),
            ("Product license", self.find_license_information, "central-license.json"),
            ("Networking", self.find_network_interfaces, "network-interfaces.json"),
            ("DNS", self.find_nameservers, "network-nameservers.json"),
            ("Hypervisors", self.find_hypervisors, "hypervisors.json"),
            ("Replication targets", self.find_replication_targets, "replication-targets.json"),
            ("Datasets", self.find_zfs_datasets, "datasets.json"),
            ("Protected objects", self.find_zfs_datasets_metadata, "datasets-metadata.json"),
            ("Virtual ecosystem", self.find_vmware_objects, "virtual-ecosystem.json"),
            ("Protection schedules", self.find_protection_schedules, "protection-schedules.json"),
            ("SSH key-pairs", self.find_ssh_keys, "ssh-keys.json"),
            ("Email alerts", self.find_alerter, "email-alerts.json"),
        ]

        # Iterate over the metadata items and process each one
        for description, function, filename in metadata_items:
            print("  {description} ...".format(description=description))
            data = function()

            file_path = os.path.join(target_dir, filename)
            with open(file_path, "w", encoding="utf8") as f:
                json.dump(data, f, ensure_ascii=False, indent=2)

        self.report_shares_usage()
        self.report_vms_per_dataset()
        print("\nCompleted collection of SynetoOS4 migration metadata")
        return self._serial

    def find_survey_script_version(self):
        result = {
            "version": VERSION,
            "job_uuid": str(uuid.uuid4()),
            "script_path": os.path.abspath(__file__),
        }
        try:
            result["user"] = getpass.getuser()
        except Exception as e:
            error_type = type(e).__name__
            print(
                "Failed to validate current user error={error} message={message}".format(
                    error=error_type, message=str(e)
                )
            )
            result.update({"user_error": str(e) or error_type, "user_status": "Error"})

        try:
            with open(__file__, "rt", encoding="utf8") as f:
                content = f.read()
                script_hash = hashlib.sha256(content.encode("utf8")).hexdigest()
                result.update({"script_hash": script_hash, "script_content": content})
        except Exception as e:
            error_type = type(e).__name__
            print(
                "Failed to validate survey script error={error} message={message}".format(
                    error=error_type, message=str(e)
                )
            )
            result.update({"script_error": str(e) or error_type, "script_status": "Error"})

        return result

    def find_firmware_version(self, skip_check: bool = False) -> dict:
        result = {"serial_number": "UNKNOWN", "syneto_version": "4.X.Y.0000"}

        try:
            lego_output = subprocess.check_output(
                ["config", "firm", "show"],
                shell=False,
                stderr=subprocess.DEVNULL,
                timeout=10.0,
            ).decode("utf-8")
            for line in lego_output.splitlines():
                if "Syneto Storage" in line:
                    result["syneto_version"] = line.split(" ")[-1]
                elif "SynetoOS Version:" in line:
                    result["syneto_version"] = line.split(" ")[-1]
                else:
                    key, value = line.split(":")[:2]
                    result[key.strip().lower().replace(" ", "_")] = value.strip()

        except subprocess.TimeoutExpired:
            print("Failed to check SynetoOS version, upgrade to OS5 may fail")

        self._serial = result.get("serial_number")
        self._version = result.get("syneto_version")
        if not skip_check and not self._version.startswith("4."):
            raise VersionError(
                "Unsupported SynetoOS version: {version}. This tool is designed for SynetoOS 4.x".format(
                    version=self._version
                )
            )

        return result

    def find_license_information(self) -> dict:
        try:
            subprocess.check_call(
                ["/usr/sbin/invoke-lego-api", "Central\Facade", "sync", "-J"],
                shell=False,
                stderr=subprocess.DEVNULL,
                timeout=90.0,
            )
        except subprocess.CalledProcessError as e:
            if self.check_license:
                print("Failed to sync central license code={code}".format(code=e.returncode))

        try:
            with open("/var/storage/central.product", "r") as product_file:
                product = json.loads(product_file.read())
            with open("/var/storage/central.token", "r") as token_file:
                token = token_file.read()

            if self.check_license and not product.get("isUpgradableToOS5"):
                print("License validation failed, to complete the upgrade, please purchase an OS5 upgrade license")

            return {"product": product, "token": token}

        except FileNotFoundError as e:
            if self.check_license:
                print("Failed validate license, upgrade to OS5 may fail message=license file not found")
            return {"product": {"status": "error", "message": str(e)}, "token": None}

        except Exception as e:
            error_type = type(e).__name__
            print(
                "Failed validate license, upgrade to OS5 may fail error={error} message={message}".format(
                    error=error_type, message=str(e)
                )
            )
            return {"product": {"status": "error", "message": str(e) or error_type}, "token": None}

    def find_network_interfaces(self) -> dict:
        lego_output = subprocess.check_output(
            ["/usr/sbin/invoke-lego-api", "Networking_Repository_Interface", "findAll", "-J"], shell=False
        ).decode("utf-8")
        return json.loads(lego_output)

    def find_nameservers(self) -> dict:
        with open("/etc/resolv.conf", "r") as resolv_conf:
            nameservers = [line.split() for line in resolv_conf.readlines() if line.startswith("nameserver")]
            return {"nameservers": [item for sublist in nameservers for item in sublist if item != "nameserver"]}

    def find_hypervisors(self) -> list:
        # Use config parser to read /var/storage/vhosts
        config = configparser.ConfigParser()
        config.read("/var/storage/vhosts")

        hypervisors = []
        for section in config.sections():
            hypervisor_info = {
                "ip": section,
                "username": config[section].get("username"),
                "password": config[section].get("password"),
                "isLocalHypervisor": config[section].getboolean("isLocalHypervisor", fallback=False),
            }
            hypervisors.append(hypervisor_info)
        self._hypervisors = hypervisors
        return hypervisors

    def find_protection_schedules(self) -> list:
        # Collect dataset protection schedules
        try:
            lego_output = subprocess.check_output(
                ["/usr/sbin/invoke-lego-api", "Backup\Facade", "getSnapshotServicesAsHash", "-J"],
                shell=False,
                timeout=90.0,
            ).decode("utf-8")
            protection_schedules = json.loads(lego_output)
            if not protection_schedules:
                return []
            return [{**schedule, "dataset": dataset} for dataset, schedule in protection_schedules.items()]
        except Exception as e:
            print(
                "Failed to collect protection schedules error={error} message={message}".format(
                    error=type(e).__name__, message=str(e)
                )
            )
            return []

    def find_replication_targets(self) -> list:
        # Collect contents of /var/storage/replication_hosts
        lego_output = subprocess.check_output(
            ["/usr/sbin/invoke-lego-api", "\Protection\ReplicationHost\ReplicationTargetFacade", "getAll", "-J"],
            shell=False,
        ).decode("utf-8")
        replication_targets = json.loads(lego_output)
        if not replication_targets:
            return []
        return [{**host, "key": key} for key, host in replication_targets.items()]

    def find_zfs_pools(self) -> list:
        zpoollist_output = subprocess.check_output(
            ["zpool", "list", "-H", "-o", "name"],
            shell=False,
        ).decode("utf-8")
        return [line for line in zpoollist_output.splitlines() if not line.startswith("rpool")]

    def get_zpool_roots(self) -> list:
        return ["/{}".format(pool) for pool in self._pools]

    def find_zfs_datasets(self) -> list:
        zfslist_output = subprocess.check_output(
            ["zfs", "list", "-H", "-o", "space,origin,canmount,mounted,nbmand,sharenfs,sharesmb,mountpoint"],
            shell=False,
        ).decode("utf-8")

        result = []
        for line in zfslist_output.splitlines():
            if line.startswith("rpool"):
                continue
            values = line.split()
            dataset, avail, used, usedsnap, usedds, usedrefreserv, usedchild, origin = values[:8]
            canmount, mounted, nbmand, sharenfs, sharesmb, mountpoint = values[8:14]
            result.append(
                {
                    "dataset": dataset,
                    "avail": avail,
                    "used": used,
                    "usedsnap": usedsnap,
                    "usedds": usedds,
                    "usedrefreserv": usedrefreserv,
                    "usedchild": usedchild,
                    "origin": origin,
                    "canmount": canmount,
                    "mounted": mounted,
                    "nbmand": nbmand,
                    "sharenfs": sharenfs,
                    "sharesmb": sharesmb,
                    "mountpoint": mountpoint,
                }
            )
        self._datasets_reporting_data = result
        return result

    def find_zfs_datasets_metadata(self) -> list:
        zfslist_output = subprocess.check_output(["zfs", "list", "-H", "-o", "name"], shell=False).decode("utf-8")
        datasets = [line.strip() for line in zfslist_output.splitlines() if not line.startswith("rpool")]

        with ThreadPoolExecutor(max_workers=4) as executor:
            return list(executor.map(_get_zfs_meta, datasets))

    def find_vmware_objects(self) -> list:
        with ThreadPoolExecutor(max_workers=4) as executor:
            result = list(executor.map(self._collect_hypervisor_config, self._hypervisors))
            self._vm_reporting_data = [vm for record in result for vm in record["virtual_machines"]]
            return result

    def _collect_hypervisor_config(self, hypervisor) -> dict:
        service = None
        try:
            service = SmartConnectNoSSL(
                host=hypervisor["ip"],
                port=443,
                user=hypervisor["username"],
                pwd=_decode_syneto_encoded(hypervisor["password"]),
                connectionPoolTimeout=45.0,
            )
            datacenters = {}
            for item in service.content.rootFolder.childEntity:
                if isinstance(item, vim.Datacenter):
                    datacenters[item.name.lower()] = item

            all_datastores = list(_enumerate_datastores(service))
            physical_hosts = []
            virtual_machines = []
            for dc_name, datacenter in datacenters.items():
                for cluster in datacenter.hostFolder.childEntity:
                    for host in cluster.host:
                        try:
                            physical_hosts.append(self._collect_host_config(dc_name, host))
                        except Exception as e:
                            host_record = {
                                "status": "error",
                                "name": host.name,
                                "moref_id": host._moId,
                                "error": "broken error={error} message={message}".format(
                                    error=type(e).__name__, message=str(e)
                                ),
                            }
                            physical_hosts.append(host_record)

                for vm in _enumerate_datacenter_vms(datacenter.vmFolder):
                    try:
                        virtual_machines.append(self._collect_vm_config(all_datastores, dc_name, vm))

                    except Exception as e:
                        vm_record = {
                            "status": "error",
                            "name": vm.name,
                            "moref_id": vm._moId,
                            "error": "broken error={error} message={message}".format(
                                error=type(e).__name__, message=str(e)
                            ),
                        }
                        virtual_machines.append(vm_record)

            datastores_config = [self._collect_datastore_config(ds) for ds in all_datastores]
            return {
                "hypervisor": hypervisor["ip"],
                "username": hypervisor["username"],
                "hosts": physical_hosts,
                "virtual_machines": virtual_machines,
                "datastores": datastores_config,
            }

        except vim.fault.InvalidLogin as e:
            message = e.msg or str(e)
            print("Failed to connect to {hypervisor}: {message}".format(hypervisor=hypervisor["ip"], message=message))
            raise ConnectionError("Failed to collect virtual machine migration data")

        except Exception as e:
            print(
                "Failed to collect VMs hypervisor={hypervisor} error={error} message={message}".format(
                    hypervisor=hypervisor["ip"], error=type(e).__name__, message=str(e)
                )
            )
            return {
                "status": "error",
                "hypervisor": hypervisor["ip"],
                "username": hypervisor["username"],
                "hosts": [],
                "virtual_machines": [],
                "error": "broken error={error} message={message}".format(error=type(e).__name__, message=str(e)),
            }

        finally:
            if service is not None:
                try:
                    Disconnect(service)
                    service._stub.DropConnections()
                except Exception as e:
                    print(
                        "Failed to disconnect hypervisor={hypervisor} error={error} message={message}".format(
                            hypervisor=hypervisor["ip"], error=type(e).__name__, message=str(e)
                        )
                    )

    def _collect_host_config(self, dc_name, host):
        host_system_info = host.hardware.systemInfo
        hardware = host.summary.hardware

        return {
            "name": host.name,
            "moref_id": host._moId,
            "uuid": host_system_info.uuid,
            "model": host_system_info.model,
            "vendor": host_system_info.vendor,
            "cpu": {
                "model": hardware.cpuModel,
                "frequency_hz": hardware.cpuMhz * (1000**2),
                "core_count": hardware.numCpuCores,
                "sockets": hardware.numCpuPkgs,
            },
            "memory": hardware.memorySize,
        }

    def _collect_vm_config(self, all_datastores, dc_name, vm):
        vm_config = vm.config
        vm_config_hardware = vm_config.hardware

        disk_deices = []
        cdrom_deices = []
        network_devices = []
        capacity_in_bytes = 0
        for device in vm_config_hardware.device:
            if isinstance(device, vim.vm.device.VirtualDisk):
                device_record = self._collect_vdisk_config(device)
                capacity_in_bytes += device_record.get("storage_capacity", 0)
                disk_deices.append(device_record)

            elif isinstance(device, vim.vm.device.VirtualCdrom):
                cdrom_deices.append(self._collect_vcdrom_config(device))
                continue

            elif isinstance(device, vim.vm.device.VirtualEthernetCard):
                network_devices.append(self._collect_ethernet_config(device))
                continue

        vm_runtime = vm.runtime
        vm_runtime_host = vm_runtime.host
        host_system_info = vm_runtime_host.hardware.systemInfo

        datastores = self._find_datastores(all_datastores, vm_config.datastoreUrl)
        return {
            "status": "ok",
            "datacenter": dc_name,
            "host_name": vm_runtime_host.name,
            "host_uuid": host_system_info.uuid,
            "name": vm.name,
            "uuid": vm_config.uuid,
            "moref_id": vm._moId,
            "vmx_path": vm_config.files.vmPathName,
            "datastores": [self._collect_datastore_config(ds) for ds in datastores],
            "cpu": {
                "core_count": vm_config_hardware.numCPU,
                "cores_per_socket": vm_config_hardware.numCoresPerSocket,
            },
            "memory": int(vm_config_hardware.memoryMB) * 1024 * 1024,
            "storage": {
                "capacity_in_bytes": capacity_in_bytes,
                "disk_devices": disk_deices,
                "cdrom_devices": cdrom_deices,
            },
            "network": {"ethernet_devices": network_devices},
        }

    def _collect_vdisk_config(self, device):
        backing = device.backing
        device_info = device.deviceInfo
        datastore = backing.datastore

        return {
            "device_info_label": device_info.label,
            "storage_capacity": device.capacityInBytes,
            "filepath": get_vm_disk_root_filename(backing),
            "backing_datastore": self._collect_datastore_config(datastore) if datastore else None,
            "backing_disk_mode": backing.diskMode,
            "backing_type": type(backing).__name__,
        }

    def _collect_vcdrom_config(self, device):
        backing = device.backing
        device_info = device.deviceInfo
        record = {
            "device_info_label": device_info.label,
            "device_info_summary": device_info.summary,
            "backing_type": type(backing).__name__,
        }
        if type(backing) is vim.vm.device.VirtualCdrom.RemoteAtapiBackingInfo:
            record["backing_device_name"] = backing.deviceName

        elif type(backing) is vim.vm.device.VirtualCdrom.IsoBackingInfo:
            datastore = backing.datastore
            record["backing_file_name"] = backing.fileName
            record["backing_datastore"] = self._collect_datastore_config(datastore) if datastore else None

        return record

    def _collect_datastore_config(self, datastore):
        if not datastore:
            raise SurveyError("Datastore must not be None")
        ds_record = {
            "moref_id": datastore._moId,
            "name": datastore.name,
            "key": str(datastore),
            "type": datastore.summary.type,
            "capacity": datastore.summary.capacity,
            "info_type": type(datastore.info).__name__,
            "accessible ": datastore.summary.accessible,
            "url": datastore.summary.url,
            "hosts": [self._collect_datastore_mount(hst_mnt) for hst_mnt in datastore.host],
        }

        if type(datastore.info) is vim.host.NasDatastoreInfo:
            ds_nas = datastore.info.nas
            ds_record["remote_host"] = ds_nas.remoteHost
            ds_record["remote_path"] = ds_nas.remotePath

        elif type(datastore.info) is vim.host.VmfsDatastoreInfo:
            ds_vmfs = datastore.info.vmfs
            ds_record["vmfs_local"] = ds_vmfs.local
            ds_record["vmfs_ssd"] = ds_vmfs.ssd
            ds_record["vmfs_uuid"] = ds_vmfs.uuid
            ds_record["vmfs_extents"] = _get_vmfs_disk_extents(ds_vmfs)
        else:
            # TODO: replace, make it work with iSCSI
            print(datastore.info)

        return ds_record

    def _collect_datastore_mount(self, ds_host_mount):
        return {
            "host": str(ds_host_mount.key),
            "path": ds_host_mount.mountInfo.path,
            "accessMode": ds_host_mount.mountInfo.accessMode,
            "mounted": ds_host_mount.mountInfo.mounted,
            "accessible": ds_host_mount.mountInfo.accessible,
            "inaccessibleReason": ds_host_mount.mountInfo.inaccessibleReason,
        }

    def _collect_ethernet_config(self, device):
        info = device.deviceInfo
        connectable = device.connectable
        backing = device.backing
        return {
            "label": info.label,
            "summary": info.summary,
            "mac_address": device.macAddress,
            "backing_device": backing.deviceName,
            "backing_network": str(backing.network).strip("'"),
            "start_connected": connectable.startConnected,
            "connected": connectable.connected,
            "status": connectable.status,
        }

    def _find_datastores(self, all_datastores, datastore_url_refs):
        results = []

        for ds_ref in datastore_url_refs:
            name = ds_ref.name
            dsurl = _normalize_ds_url(ds_ref.url)

            matches = [
                datastore
                for datastore in all_datastores
                if datastore.name == name and _normalize_ds_url(datastore.info.url) == dsurl
            ]
            if len(matches) > 1:
                print("Multiple datastores match name={name}".format(name=name))
            elif len(matches) == 0:
                matches = [datastore for datastore in all_datastores if datastore.name == name]
                if len(matches) == 0:
                    print("No datastores match name={name} path={dsurl}".format(name=name, dsurl=dsurl))
                else:
                    print(
                        "No datastores match both name={name} and path={dsurl}, some only match the name:".format(
                            name=name, dsurl=dsurl
                        )
                    )
                    for ds in matches:
                        print(ds.info)

            results.extend(matches)
        return results

    def find_ssh_keys(self) -> dict:
        # TODO: collect contents ssh key pairs and authorized_keys for admin, zfssnap, support
        return {}

    def find_alerter(self) -> dict:
        # TODO: collect contents of mail_event_recipients.conf, mail_sender_address.conf and mailer.ini
        return {}

    def report_shares_usage(self):
        if not self._datasets_reporting_data:
            return
        shares = []
        volumes = []
        others = []
        for record in self._datasets_reporting_data:
            dataset = record["dataset"]
            sharesmb = record["sharesmb"]
            used = record["used"]

            parts = dataset.split("/")
            if len(parts) <= 2:
                continue
            pool_name = parts[0]
            formatted_pool_name = pool_name if len(pool_name) <= 10 else (pool_name[:7] + "...")
            if self._is_share_dataset(dataset) and len(parts) >= 4:
                share_type, share_name = parts[2], "/".join(parts[3:])
                shares.append(
                    "{type:5}| {pool:10} | {used:9} | {share:20}".format(
                        type=share_type.upper(), pool=formatted_pool_name, used=used, share=share_name
                    )
                )
            elif self._is_volume_dataset(dataset) and len(parts) >= 3:
                volume_name = "/".join(parts[2:])
                volumes.append(
                    "{pool:10} | {used:9} | {volume:20}".format(pool=formatted_pool_name, used=used, volume=volume_name)
                )
            elif sharesmb and sharesmb.lower() != "off":
                others.append(
                    "{pool:10} | {used:9} | {dataset:20}".format(pool=formatted_pool_name, used=used, dataset=dataset)
                )

            os.path.split(dataset)

        if not others:
            print("Found {shares} shares and {volumes} volumes".format(shares=len(shares), volumes=len(volumes)))
        else:
            print(
                "Found {shares} shares, {volumes} volumes and {others} other SMB configurations".format(
                    shares=len(shares), volumes=len(volumes), others=len(others)
                )
            )
        if shares:
            print("\nSMB/AFP Shares")
            print(
                "ATTENTION: Shares are present on this appliance. For migration to SynetoOS5, please follow "
                "the guidelines. Datasets with Shares are available in the commandline at:"
            )
            print("Type | Pool Name  | Used Size | Share name")
            print("\n".join(shares))

        if volumes:
            print("\niSCSI Volumes")
            print(
                "ATTENTION: iSCSI volumes are present on this appliance. For migration to SynetoOS5, please follow "
                "the guidelines. iSCSI volumes are available in the commandline at:"
            )
            print("Pool Name  | Used Storage | Volume name")
            print("\n".join(volumes))

        if others:
            print("\nOther SMB configurations")
            print(
                "ATTENTION: Datasets with custom share config are present on this appliance. For migration to SynetoOS5, please follow "
                "the guidelines. These datasets are available in the commandline at:"
            )
            print("Pool Name  | Used Storage | Dataset")
            print("\n".join(others))

    def _get_vms_grouped_by_datastore(self):
        datastore_vms = {}
        if not self._vm_reporting_data:
            return datastore_vms

        for vm in self._vm_reporting_data:
            vmx_path = vm.get("vmx_path", "N/A")
            for ds in vm.get("datastores", []):
                ds_remote_path = ds.get("remote_path")
                ds_name_in_path = "[{ds_name}] ".format(ds_name=ds["name"])
                if not ds_remote_path or not vmx_path.startswith(ds_name_in_path):
                    continue
                local_vmx_path = os.path.join(ds_remote_path, vmx_path[len(ds_name_in_path) :].lstrip("/"))
                if os.path.isfile(local_vmx_path):
                    vms = datastore_vms[ds_remote_path] = datastore_vms.get(ds_remote_path, [])
                    vms.append({"local_vmx_path": local_vmx_path, "vm": vm})
        return datastore_vms

    def _is_matching(self, dataset_name, included):
        parts = dataset_name.replace("shares/afp", "shares").replace("shares/smb", "shares").split("/")
        if len(parts) <= 2:
            return False
        pool_name, storage_type = parts[:2]
        if storage_type in "backups" and len(parts) == 3:
            return False
        elif storage_type in "backups" and len(parts) >= 4:
            return self._is_matching("/".join(parts[3:]), included=included)
        elif included:
            return storage_type in included
        else:
            raise ValueError("A filter arguments is required, please use `included_types`")

    def _is_share_dataset(self, dataset_name):
        return self._is_matching(dataset_name, included=["shares"])

    def _is_volume_dataset(self, dataset_name):
        return self._is_matching(dataset_name, included=["volumes"])

    def _is_datastore(self, dataset_name):
        return self._is_matching(dataset_name, included=["datastores"])

    def report_vms_per_dataset(self):
        datastore_vms = self._get_vms_grouped_by_datastore()

        for record in self._datasets_reporting_data:
            if not self._is_datastore(record["dataset"]):
                continue
            ds_mountpoint = record["mountpoint"]
            datastore_vms[ds_mountpoint] = datastore_vms.get(ds_mountpoint, [])

        datasets_without_vms = [ds_mountpoint for ds_mountpoint, records in datastore_vms.items() if not records]
        datasets_with_one_vm = [records[0] for ds_mountpoint, records in datastore_vms.items() if len(records) == 1]
        datasets_with_many_vms = {
            ds_mountpoint: records for ds_mountpoint, records in datastore_vms.items() if len(records) > 1
        }

        print("\n{nominal} virtual machines can be migrated automatically.".format(nominal=len(datasets_with_one_vm)))
        # datasets_with_many_vms = datastore_vms
        if datasets_with_many_vms:
            print(
                "ATTENTION: {many_vms} datasets store more than one virtual machine.".format(
                    many_vms=len(datasets_with_many_vms)
                )
            )
            print(
                "ACTION: Please move the data for each virtual machine to a separate dataset and restart "
                "the migration. A dataset must host only one virtual machine."
            )
            for ds_mountpoint, records in datasets_with_many_vms.items():
                print("Dataset {mountpoint}:".format(mountpoint=ds_mountpoint))
                for record in records:
                    print("    VM: {name} - {path}".format(name=record["vm"]["name"], path=record["local_vmx_path"]))

        if datasets_without_vms:
            print(
                "ATTENTION: {no_vms} datasets do not have any registered virtual machines.".format(
                    no_vms=len(datasets_without_vms)
                )
            )
            print(
                "For migration to SynetoOS5, please follow the guidelines. "
                "These datasets are available in the commandline at:"
            )
            print("\n".join(datasets_without_vms))


def _normalize_ds_url(ds_url):
    # datastoreUrl.url entries are of the forms:
    # - /vmfs/volumes/123456-789abcd/
    # - /vmfs/volumes/123456-789abcd... (without trailing slash)
    # - ds:///vmfs/volumes/123456-789abcd/ (with a ds:// prefix)
    # in some places there is a trailing "/" in some others there's none
    # in some place there is a ds:// prefix, others not (on the same server)
    # ... so we normalize it
    ds_url = ds_url.strip(" ")
    prefix = "ds://"
    if ds_url.lower().startswith(prefix):
        ds_url = ds_url[len(prefix) :]
    return ds_url.rstrip("/")


def get_vm_disk_root_filename(disk_device_backing):
    actual_backing = disk_device_backing

    while actual_backing.parent:
        actual_backing = actual_backing.parent

    return actual_backing.fileName


def _get_vmfs_disk_extents(vmfs_info):
    extents = []
    for extent in vmfs_info.extent:
        extents.append(
            {
                "type": type(extent).__name__,
                "disk_name": extent.diskName,
                "partition": extent.partition,
            }
        )
    return extents


def _enumerate_datacenter_vms(vm_folder: "vim.Folder", folder: str = None):
    # Get datacenter virtual machines recursively yield each vm object
    folder = folder or "/"

    for vm_or_folder in vm_folder.childEntity:
        if isinstance(vm_or_folder, vim.VirtualMachine):
            yield vm_or_folder
        elif isinstance(vm_or_folder, vim.Folder):
            sub_folder = "{folder}{subfolder}/".format(folder=folder, subfolder=vm_or_folder.name)
            for vm in _enumerate_datacenter_vms(vm_or_folder, sub_folder):
                yield vm


def _enumerate_datastores(service):
    for datacenter in service.content.rootFolder.childEntity:
        if hasattr(datacenter, "datastoreFolder"):
            datastore_folder = datacenter.datastoreFolder
            for datastore in datastore_folder.childEntity:
                yield datastore


def _decode_syneto_encoded(encoded):
    prefix = "syneto-encoded-"
    if not encoded.startswith(prefix):
        raise ConnectionError("Invalid syneto-encoded value")
    b64value = encoded[len(prefix) :]
    decoded = base64.b64decode(b64value).decode("utf-8")
    return decoded


def _get_zfs_meta(dataset: str) -> dict:
    record = {"dataset": dataset, "metadata": {}}

    try:
        zfsmeta_output = subprocess.check_output(["/usr/sbin/zfs-meta", "get", dataset], shell=False).decode("utf-8")
        zfsmeta_output = zfsmeta_output.strip()
        if zfsmeta_output and zfsmeta_output == "[]":
            return record
        record["metadata"] = json.loads(zfsmeta_output)
        return record

    except Exception as e:
        print(
            "Failed zfs-meta get {dataset}, error={error} message={message}".format(
                dataset=dataset, error=type(e).__name__, message=str(e)
            )
        )
        return record


def main():
    collect_migration_data(check_license="--skip-license-check" not in sys.argv)


if __name__ == "__main__":
    main()
